import os
import shlex
import subprocess
import tempfile
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import IO, Any, TextIO, Union


@dataclass(frozen=True, kw_only=True)
class CommandOutput:
    stdout: str
    stderr: str


def run_cmd(
    cmd: Union[str, list[str]],
    working_dir: Path | None = None,
    additional_env: dict[str, str] = {},
    **kwargs: Any,  # https://github.com/python/mypy/issues/8772
) -> CommandOutput:
    if working_dir is None:
        working_dir = Path(os.getcwd())
    env = os.environ.copy()
    env.update(additional_env)

    if isinstance(cmd, list):
        cmd = " ".join(cmd)
    output = subprocess.run(
        shlex.split(cmd.replace('"', '\\"')),
        cwd=str(working_dir),
        check=True,
        env=env,
        capture_output=True,
        **kwargs,
    )

    return CommandOutput(
        stdout=output.stdout.decode("utf-8").strip(),
        stderr=output.stderr.decode("utf-8").strip(),
    )


def run_cmd_async(
    cmd: Union[str, list[str]],
    working_dir: Path | None = None,
    additional_env: dict[str, str] = {},
    stdout: IO[str] | int | None = subprocess.PIPE,
    stderr: IO[str] | int | None = subprocess.PIPE,
    **kwargs: Any,
) -> subprocess.Popen[Any]:
    if working_dir is None:
        working_dir = Path(os.getcwd())
    env = os.environ.copy()
    env.update(additional_env)

    if isinstance(cmd, list):
        cmd = " ".join(cmd)

    return subprocess.Popen(
        shlex.split(cmd.replace('"', '\\"')),
        cwd=str(working_dir),
        env=env,
        stdout=stdout,
        stderr=stderr,
        **kwargs,
    )


def run_cmd_to_logfile(
    cmd: Union[str, list[str]],
    log_file: TextIO | None = None,
    working_dir: Path | None = None,
    additional_env: dict[str, str] = {},
) -> None:
    if working_dir is None:
        working_dir = Path(os.getcwd())
    env = os.environ.copy()
    env.update(additional_env)

    if isinstance(cmd, list):
        cmd = " ".join(cmd)

    subprocess.run(
        shlex.split(cmd),
        cwd=working_dir,
        check=True,
        stdout=log_file,
        stderr=subprocess.STDOUT,
        env=env,
        capture_output=False,
    )


class TempDirEnv:
    def __init__(self, change_dir: bool = False) -> None:
        self.td: tempfile.TemporaryDirectory[str]
        self.old_dir: Path

        self.chdir = change_dir

    def __enter__(self) -> Path:
        self.td = tempfile.TemporaryDirectory()
        tempfile.tempdir = self.td.name
        tmpdir_path = Path(self.td.name)
        if self.chdir:
            self.old_dir = Path(os.getcwd()).absolute()
            os.chdir(tmpdir_path)
        return tmpdir_path

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        if self.chdir:
            os.chdir(self.old_dir)
        tempfile.tempdir = None


def temporary_file(
    *, contents: str | None = None, suffix: str | None = None, delete: bool = True
) -> IO[bytes]:
    """Creates a named temporary file with extension
    `suffix` and writes `contents` into it.

    Args:
        contents (str):
            what to write in the temporary file
        suffix (str):
            the file's extension (e.g., ".c")
    Returns:
        tempfile.NamedTemporaryFile:
            a temporary file that is automatically deleted when the object is
            garbage collected
    """
    ntf = tempfile.NamedTemporaryFile(suffix=suffix, delete=delete)
    if contents:
        with open(ntf.name, "w") as f:
            f.write(contents)
    return ntf



from pathlib import Path
from shutil import which

from diopter.compiler import (
    ASMCompilationOutput,
    CompilationSetting,
    CompilerExe,
    CompilerProject,
    ExeCompilationOutput,
    Language,
    LLVMIRCompilationOutput,
    ObjectCompilationOutput,
    Opt,
    OptLevel,
    SourceProgram,
)
from diopter.utils import run_cmd, temporary_file


def test_compiler_exe_from_path() -> None:
    for v in [14, 15, 16]:
        clang_path = Path(f"clang-{v}")
        if not clang_path.exists():
            continue
        clang = CompilerExe.from_path(clang_path)
        assert clang.exe == Path(clang_path)
        assert f"{v}." in clang.revision
        assert clang.project == CompilerProject.LLVM


def test_get_asm_from_program() -> None:
    input_code = "int foo(int a){ return a + 1; }"
    program = SourceProgram(code=input_code, language=Language.C)
    compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2, flags=("-m32",))
    asm = cs.compile_program(program, ASMCompilationOutput()).output.read()

    tf = temporary_file(contents=program.code, suffix=program.get_file_suffix())
    result = run_cmd(f"gcc {tf.name} -mno-red-zone -o /dev/stdout -O2 -m32 -S")
    asm_manual = result.stdout

    def canonicalize(asm: str) -> str:
        return "\n".join(
            line for line in asm.splitlines() if ".file" not in line
        ).strip()

    assert canonicalize(asm) == canonicalize(asm_manual)


def strip_and_read_binary(path: Path) -> bytes:
    run_cmd(f"strip {path}")
    with open(str(path), "rb") as f:
        return f.read()


def test_compile_to_object() -> None:
    input_code = "int foo(int a){ return a + 1; }"
    program = SourceProgram(code=input_code, language=Language.C)
    compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2)
    res = cs.compile_program(
        program,
        ObjectCompilationOutput(),
    )
    res.output.strip_symbols()
    object1 = res.output.read()

    code_file = temporary_file(contents=program.code, suffix=program.get_file_suffix())
    object_file2 = temporary_file(contents="", suffix=".o")
    cmd = f"gcc {code_file.name} -o {object_file2.name} -O2 -c"
    run_cmd(cmd)

    object2 = strip_and_read_binary(Path(object_file2.name))

    assert object1 == object2


def test_compile_to_object_cpp() -> None:
    input_code = """#include <iostream>
                    void foo(int a){ std::cout<< a; }
                 """
    program = SourceProgram(code=input_code, language=Language.CPP)
    compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2)
    res = cs.compile_program(
        program,
        ObjectCompilationOutput(),
    )

    res.output.strip_symbols()
    object1 = res.output.read()

    object_file2 = temporary_file(contents="", suffix=".o")
    code_file = temporary_file(contents=program.code, suffix=program.get_file_suffix())
    cmd = f"g++ {code_file.name} -o {object_file2.name} -O2 -c"
    run_cmd(cmd)

    object2 = strip_and_read_binary(Path(object_file2.name))

    assert object1 == object2


def test_compile_to_exec() -> None:
    input_code = "int foo(int a){ return a + 1; } int main(){return foo(1);}"
    program = SourceProgram(code=input_code, language=Language.C)
    compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2)
    res = cs.compile_program(
        program,
        ExeCompilationOutput(),
    )
    assert which(res.output.filename)
    res.output.strip_symbols()
    exe1 = res.output.read()

    exe_file2 = temporary_file(contents="", suffix=".exe")
    code_file = temporary_file(contents=program.code, suffix=program.get_file_suffix())
    cmd = f"gcc {code_file.name} -o {exe_file2.name} -O2 "
    run_cmd(cmd)

    assert which(exe_file2.name)
    exe2 = strip_and_read_binary(Path(exe_file2.name))

    assert exe1 == exe2


def test_preprocess() -> None:
    input_code = """
    #define MACRO1 4
    int foo(){
        return MACRO1 + MACRO2;
    }
    """
    program = SourceProgram(code=input_code, language=Language.C)
    compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2)
    pp_code = cs.preprocess_program(program, False, ("-DMACRO2=33",)).code
    assert "".join(pp_code.split()) == "".join(
        "int foo(){ return 4 + 33; }".split()
    ), pp_code


def test_exe_run() -> None:
    input_code = """
    #include <stdio.h>
    void foo(int argc){
        printf("%d \\n", argc);
    }
    int main(int argc, char* argv[]){
        foo(argc);
    }
    """
    program = SourceProgram(code=input_code, language=Language.C)
    compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2)
    res = cs.compile_program(
        program,
        ExeCompilationOutput(),
    )
    output = res.output.run()
    assert output.stdout.strip() == "1"
    output = res.output.run(("asdf",))
    assert output.stdout.strip() == "2"
    output = res.output.run(("asdf", "fff"))
    assert output.stdout.strip() == "3"


def test_async_compile() -> None:
    input_code = "int foo(int a){ return a + 1; }"
    program = SourceProgram(code=input_code, language=Language.C)
    compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2)
    res1 = cs.compile_program(
        program,
        ObjectCompilationOutput(),
    )
    res1.output.strip_symbols()
    object1 = res1.output.read()

    res2 = cs.compile_program_async(
        program,
        ObjectCompilationOutput(),
    ).result(8)

    res2.output.strip_symbols()
    object2 = res2.output.read()

    assert object1 == object2

    async_res3 = cs.compile_program_async(
        program,
        ObjectCompilationOutput(),
    )
    # calling wait should not change the result
    async_res3.wait(8)
    res3 = async_res3.result(8)
    res3.output.strip_symbols()
    object3 = res3.output.read()

    assert object2 == object3


def test_link() -> None:
    input_code1 = "int foo(int a){ return a + 1; }"
    input_code2 = """
                  int foo(int);
                  int bar(int a){ return foo(a + 1); }
                  """
    input_code3 = """
                  int bar(int);
                  int main(int argc, char* argv){ return bar(argc); }
                  """
    program1 = SourceProgram(code=input_code1, language=Language.C)
    program2 = SourceProgram(code=input_code2, language=Language.C)
    program3 = SourceProgram(code=input_code3, language=Language.C)
    compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2)
    res1 = cs.compile_program(
        program1,
        ObjectCompilationOutput(),
    )
    res2 = cs.compile_program(
        program2,
        ObjectCompilationOutput(),
    )
    res3 = cs.compile_program(
        program3,
        ObjectCompilationOutput(),
    )
    exe_res1 = cs.link_objects(
        (res1.output, res2.output, res3.output), ExeCompilationOutput()
    )
    exe_res1.output.strip_symbols()
    exe_res2 = cs.link_objects_async(
        (res1.output, res2.output, res3.output), ExeCompilationOutput()
    ).result()
    exe_res2.output.strip_symbols()

    code_file1 = temporary_file(
        contents=program1.code, suffix=program1.get_file_suffix()
    )
    code_file2 = temporary_file(
        contents=program2.code, suffix=program2.get_file_suffix()
    )
    code_file3 = temporary_file(
        contents=program3.code, suffix=program3.get_file_suffix()
    )
    object_file1 = temporary_file(contents="", suffix=".o")
    object_file2 = temporary_file(contents="", suffix=".o")
    object_file3 = temporary_file(contents="", suffix=".o")
    run_cmd(f"{compiler.exe} -c -O2 {code_file1.name} -o {object_file1.name}")
    run_cmd(f"{compiler.exe} -c -O2 {code_file2.name} -o {object_file2.name}")
    run_cmd(f"{compiler.exe} -c -O2 {code_file3.name} -o {object_file3.name}")

    exe3_file = temporary_file(contents="", suffix=".exe")
    run_cmd(
        f"{compiler.exe} -O2 {object_file1.name} {object_file2.name} "
        f"{object_file3.name} -o {exe3_file.name}"
    )

    assert which(exe3_file.name)
    exe3 = strip_and_read_binary(Path(exe3_file.name))
    assert exe3 == exe_res1.output.read()
    assert exe3 == exe_res2.output.read()


def get_opt() -> Opt:
    for suffix in ["", "-13", "-14", "-15", "-16", "-17"]:
        opt_path = Path(f"opt{suffix}")
        if not which(opt_path):
            continue
        return Opt.from_path(opt_path)
    assert False, "Could not find opt exe"


def test_opt() -> None:
    input_code = """
    int foo(int a){
        return a + 1;
    }
    """
    program = SourceProgram(code=input_code, language=Language.C)
    clang = CompilerExe.get_system_clang()
    cs = CompilationSetting(compiler=clang, opt_level=OptLevel.O0)
    res = cs.compile_program(program, LLVMIRCompilationOutput())
    llvm_ir = res.output.read()
    opt = get_opt()
    opt_res = opt.run_on_input(res.output.filename, ("-S",))
    assert llvm_ir.strip() != opt_res.stdout.strip()



""" A sanitizer used to check programs for potential correctness issues

The following checks are currently supported: checking for compiler warnings,
checking a program with clang's undefined behaviour and address sanitizers,
checking a program with CompCert (ccomp).

Example:

program : SourceProgram = ...
sanitizer = Sanitizer()
if not sanitizer.sanitize(program):
    # the program is broken
"""
from __future__ import annotations

import subprocess
from dataclasses import dataclass
from pathlib import Path

from diopter.compiler import (
    CComp,
    CompilationSetting,
    CompileError,
    CompilerExe,
    ExeCompilationOutput,
    ObjectCompilationOutput,
    OptLevel,
    SourceProgram,
)
from diopter.utils import TempDirEnv, run_cmd


@dataclass(frozen=True, kw_only=True)
class SanitizationResult:
    """Why did the sanitizer fail on a program?"""

    check_warnings_failed: bool = False
    ub_address_sanitizer_failed: bool = False
    ccomp_failed: bool = False
    timeout: bool = False

    def __bool__(self) -> bool:
        """True indicates that sanitization was successful"""
        return not (
            self.check_warnings_failed
            or self.ub_address_sanitizer_failed
            or self.ccomp_failed
            or self.timeout
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, bool):
            return self.__bool__() == other
        if not isinstance(other, SanitizationResult):
            return NotImplemented
        return self == other

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)


class Sanitizer:
    """A wrapper of various sanitization methods.

    Sanitizer tests an input program using  various checks
    to rule of obvious ways a program can be broken:
    - compiler warnings
    - address and undefined behavior sanitizers
    - CompCert

    Attributes:
        gcc (CompilerExe):
            gcc used for checking compiler warnings
        clang (CompilerExe):
            clang used for checking compiler warnings and ub/address sanitizers result
        ccomp (CComp | None):
            CompCert used for validating the program
        checked_warnings (tuple[str,...] | None):
            the warnings whose presence to check
        use_ub_address_sanitizer (bool):
            whether Sanitizer.sanitize should use clang's ub and address sanitizers
        check_warnings_opt_level (OptLevel):
            optimization level used when checking for warnings
        sanitizer_opt_level (OptLevel):
            optimization level used when running sanitizers
        compilation_timeout (int):
            seconds to wait before aborting when compiling the program
        execution_timeout (int):
            seconds to wait before aborting when executing the program (ub/asan)
        ccomp_timeout  (int):
            seconds to wait before aborting when interpreting the program with ccomp
        debug (bool):
            if True then additional info is printed when sanitizing programs
    """

    default_warnings = (
        "cast to smaller integer type",
        "conversions than data arguments",
        "incompatible redeclaration",
        "ordered comparison between pointer",
        "eliding middle term",
        "end of non-void function",
        "invalid in C99",
        "specifies type",
        "should return a value",
        "uninitialized",
        "incompatible pointer to",
        "incompatible integer to",
        "comparison of distinct pointer types",
        "type specifier missing",
        "uninitialized",
        "Wimplicit-int",
        "division by zero",
        "without a cast",
        "control reaches end",
        "return type defaults",
        "cast from pointer to integer",
        "useless type name in empty declaration",
        "no semicolon at end",
        "type defaults to",
        "too few arguments for format",
        "incompatible pointer",
        "ordered comparison of pointer with integer",
        "declaration does not declare anything",
        "expects type",
        "comparison of distinct pointer types",
        "pointer from integer",
        "incompatible implicit",
        "excess elements in struct initializer",
        "comparison between pointer and integer",
        "return type of ‘main’ is not ‘int’",
        "return type of 'main' is not 'int'",
        "past the end of the array",
        "no return statement in function returning non-void",
        "undefined behavior",
    )

    def __init__(
        self,
        *,
        check_warnings: bool = True,
        use_ub_address_sanitizer: bool = True,
        use_ccomp_if_available: bool = True,
        gcc: CompilerExe | None = None,
        clang: CompilerExe | None = None,
        ccomp: CComp | None = None,
        check_warnings_opt_level: OptLevel = OptLevel.O3,
        sanitizer_opt_level: OptLevel = OptLevel.O0,
        checked_warnings: tuple[str, ...] | None = None,
        compilation_timeout: int = 8,
        execution_timeout: int = 4,
        ccomp_timeout: int = 16,
        debug: bool = False,
    ):
        """
        Args:
            check_warnings (bool):
                if True gcc's and clang's outputs will be used to
                filter out cases with Sanitizer.default_warnings
            use_ub_address_sanitizer (bool):
                whether to use clang's undefined behavior and address sanitizers
            use_ccomp_if_available (bool | None):
                if ccomp should be used, if ccomp is not None this argument is ignored
            gcc (CompilerExe | None):
                the gcc executable to use, if not provided
                CompilerExe.get_system_gcc will be used
            clang (CompilerExe | None):
                the clang executable to use, if not provided
                CompilerExe.get_system_clang will be used
            ccomp (CComp | None):
                the ccomp executable to use, if not provided and use_ccomp
                is True CComp.get_system_ccomp will be used if available
            check_warnings_opt_level (OptLevel):
                which optimization level to use when checking
                for compiler warnings
            sanitizer_opt_level (OptLevel):
                which optimization level to use when checking
                for ub/address sanitizer issues
            checked_warnings (tuple[str,...] | None):
                if not None implies check_warnings = True and will
                be used instead of Sanitizer.default_warnings
            compilation_timeout (int):
                after how many seconds to abort a compilation and fail
            execution_timeout (int):
                after how many seconds to abort an execution and fail
                (relevant for use_ub_sanitizer)
            ccomp_timeout (int):
                after how many seconds to abort interpreting with ccomp and fail
            debug (bool):
                if True then additional info is printed when sanitizing programs
        """
        self.gcc = gcc if gcc else CompilerExe.get_system_gcc()
        self.clang = clang if clang else CompilerExe.get_system_clang()
        self.ccomp = ccomp
        if use_ccomp_if_available and not self.ccomp:
            self.ccomp = CComp.get_system_ccomp()
        if checked_warnings:
            self.checked_warnings = checked_warnings
        elif check_warnings:
            self.checked_warnings = Sanitizer.default_warnings
        self.use_ub_address_sanitizer = use_ub_address_sanitizer
        self.check_warnings_opt_level = check_warnings_opt_level
        self.sanitizer_opt_level = sanitizer_opt_level
        self.compilation_timeout = compilation_timeout
        self.execution_timeout = execution_timeout
        self.ccomp_timeout = ccomp_timeout
        self.debug = debug

    def check_for_compiler_warnings(
        self, program: SourceProgram, debug: bool = False
    ) -> SanitizationResult:
        """Checks the program for compiler warnings.

        Compiles program with self.gcc and self.clang and reports
        if any of self.checked_warnings is present in their outputs
        (or if the compilation timed out).

        Args:
            program (SourceProgram):
                The program to check.
            debug (bool):
                If true, additional info will be printed in case of failure

        Returns:
            SanitizationResult:
                whether the program failed sanitization or not.
        """

        def check_warnings_impl(comp: CompilationSetting) -> SanitizationResult:
            try:
                result = comp.compile_program(
                    program,
                    ObjectCompilationOutput(Path("/dev/null")),
                    (
                        "-Wall",
                        "-Wextra",
                        "-Wpedantic",
                        "-Wno-builtin-declaration-mismatch",
                    ),
                    timeout=self.compilation_timeout,
                )
            except subprocess.TimeoutExpired:
                return SanitizationResult(timeout=True)
            except CompileError as e:
                if debug or self.debug:
                    print(e)
                return SanitizationResult(check_warnings_failed=True)
            for line in result.stdout_stderr_output.splitlines():
                for checked_warning in self.checked_warnings:
                    if checked_warning in line:
                        return SanitizationResult(check_warnings_failed=True)
            return SanitizationResult()

        with TempDirEnv():
            if not (
                gcc_result := check_warnings_impl(
                    CompilationSetting(
                        compiler=self.gcc,
                        opt_level=self.check_warnings_opt_level,
                    )
                )
            ):
                return gcc_result
            if not (
                clang_result := check_warnings_impl(
                    CompilationSetting(
                        compiler=self.clang,
                        opt_level=self.check_warnings_opt_level,
                    )
                )
            ):
                return clang_result

        return SanitizationResult()

    def check_for_ub_and_address_sanitizer_errors(
        self, program: SourceProgram, debug: bool = False
    ) -> SanitizationResult:
        """Checks the program for UB and address sanitizer errors.

        Compiles program with self.clang and -fsanitize=undefined,address, then
        it runs the checked program and reports whether it failed (or timed out).

        Args:
            program (SourceProgram):
                The program to check.
            debug (bool):
                If true, additional info will be printed in case of failure
        Returns:
            SanitizationResult:
                whether the program failed sanitization or not.
        """

        with TempDirEnv():
            # Compile program with -fsanitize=...
            try:
                result = CompilationSetting(
                    compiler=self.clang,
                    opt_level=self.sanitizer_opt_level,
                ).compile_program(
                    program,
                    ExeCompilationOutput(None),
                    (
                        "-Wall",
                        "-Wextra",
                        "-Wpedantic",
                        "-Wno-builtin-declaration-mismatch",
                        "-fsanitize=undefined,address",
                    ),
                    timeout=self.compilation_timeout,
                )
            except subprocess.TimeoutExpired:
                return SanitizationResult(timeout=True)
            except CompileError as e:
                if debug or self.debug:
                    print(e)
                return SanitizationResult(ub_address_sanitizer_failed=True)

            # Run the instrumented binary
            try:
                run_cmd(str(result.output.filename), timeout=self.execution_timeout)
            except subprocess.TimeoutExpired:
                return SanitizationResult(timeout=True)
            except subprocess.CalledProcessError as e:
                if debug or self.debug:
                    print(e.stdout)
                    print(e.stderr)
                return SanitizationResult(ub_address_sanitizer_failed=True)
            return SanitizationResult()

    def check_for_ccomp_errors(
        self, program: SourceProgram, debug: bool = False
    ) -> SanitizationResult | None:
        """Checks the program with self.ccomp if available.

        Interprets program with self.ccomp if available and reports whether
        it failed (or timed out).

        Args:
            program (SourceProgram):
                The program to check.
            debug (bool):
                If true, additional info will be printed in case of failure

        Returns:
            SanitizationResult | None:
                if self.ccomp is not None, whether the program
                failed sanitization or not.
        """
        if not self.ccomp:
            return None
        with TempDirEnv():
            try:
                if not self.ccomp.check_program(
                    program, timeout=self.ccomp_timeout, debug=debug
                ):
                    return SanitizationResult(ccomp_failed=True)
            except subprocess.TimeoutExpired:
                return SanitizationResult(timeout=True)

            return SanitizationResult()

    def sanitize(
        self, program: SourceProgram, debug: bool = False
    ) -> SanitizationResult:
        """Runs all the enabled sanitization checks.

        Runs all available sanitization checks based on self.checked_warnings,
        self.use_ub_address_sanitizer, and self.ccomp. It reports if any of them
        failed or if some check timed out.

        Args:
            program (SourceProgram):
                The program to check.
            debug (bool):
                If true, additional info will be printed in case of failure

        Returns:
            SanitizationResult:
                Whether the program failed sanitization or not.
        """

        if self.checked_warnings and not (
            check_warnings_result := self.check_for_compiler_warnings(program)
        ):
            return check_warnings_result

        if self.use_ub_address_sanitizer and not (
            sanitizer_result := self.check_for_ub_and_address_sanitizer_errors(
                program, debug=debug
            )
        ):
            return sanitizer_result

        if self.ccomp and not (
            ccomp_result := self.check_for_ccomp_errors(program, debug)
        ):
            assert ccomp_result is not None
            return ccomp_result

        return SanitizationResult()



""" A python wrapper around CReduce that works on SourcePrograms.

`Reducer` takes care of setting up and running CReduce. The reduction
script is implemented in as a `ReductionCallBack`, e.g.:

class ReductionCheck(ReductionCallback):
    def __init__(self, san: Sanitizer, ...):
        self.san = san
        ...

    def test(self, program: SourceProgram) -> bool:
        if not self.san.sanitize(program):
            return False
        # check if return is interesting
        ...

program:SourceProgram = ...  # the program we want to reduce
rprogram = reducer.reduce(program, ReductionCheck(Sanitizer(), ...))

if rprogram:
    #the reduction was successful
"""
import inspect
import logging
import os
import pickle
import subprocess
import sys
from abc import ABC, abstractmethod
from dataclasses import replace
from multiprocessing import cpu_count
from pathlib import Path
from shutil import which
from sys import stderr
from typing import TextIO

from diopter.compiler import SourceProgram
from diopter.utils import TempDirEnv, run_cmd_to_logfile


class ReductionCallback(ABC):
    """The reduction interestingness check must be
    implemented in a subclass
    """

    @abstractmethod
    def test(self, program: SourceProgram) -> bool:
        """Subclasses must implement this method.

        Args:
            program (SourceProgram):
                the reduced program, it is a copy of the original
                with it's code replaced with a reduced version
                generated by creduce

        Returns:
            (bool): is the reduced program interesting?
        """
        pass


def emit_module_imports(reduction_callback: ReductionCallback) -> str:
    """Generates all the necessary imports for the reduction script.

    Other than some stantand imports (e.g., pickle) it imports all the
    modules necessary to deserialize and run the `reduction_callback`
    """

    # Figure out the callback import:
    # from callback_module import callback_name
    callback_name = type(reduction_callback).__name__
    callback_module_path = inspect.getsourcefile(type(reduction_callback))
    assert callback_module_path
    callback_module = inspect.getmodulename(callback_module_path)

    # Also import all visible modules, this ensures that the callback
    # can be properly deserialized (pickle.load'ed) and run
    sys_path_append = "".join(f'\nsys.path.append("{p}")' for p in sys.path)

    return f"""import importlib
import pickle
import sys
from pathlib import Path
from dataclasses import replace
sys.path.insert(0, "{str(Path(callback_module_path).parent)}")
{sys_path_append}

from {callback_module} import {callback_name}
"""


def emit_call(
    reduction_callback: ReductionCallback, program: SourceProgram, code_filename: str
) -> str:
    """
    Emits the call in the reduction script.

    Serializes both the original program and the reduction call back in hex strings.
    These strings are embedded in the reduction script which pickle.load's them.

    Returns code that loads the callback and the program, reads the new code,
    and runs the callback on loaded program with the code replaced.
    """
    callback_in_hex = pickle.dumps(reduction_callback).hex()
    program_in_hex = pickle.dumps(program).hex()

    callback_load = f'callback = pickle.loads(bytes.fromhex("{callback_in_hex}"))'
    program_load = f'program = pickle.loads(bytes.fromhex("{program_in_hex}"))'

    call = "exit(not callback.test(program))"

    return f"""with open(\"{code_filename}\", \"r\") as f:
    code = f.read()
{program_load}
program = replace(program, code=code)
{callback_load}
{call}
    """


def make_interestingness_script(
    reduction_callback: ReductionCallback, program: SourceProgram, code_filename: str
) -> str:
    """
    Helper function to create a script useful for use with diopter.Reducer.
    It serializes the reduction callback into a script that checks if the code in
    code_filename is still interesting

    Args:
        reduction_callback:
            callback that will be serialized stored as a runnable script
        program:
            the original program
        code_filename:
            file that the generated script will check everytime it is run

    Returns:
        The python3 script (str)
    """
    prologue = f"#!{sys.executable}"
    return "\n".join(
        (
            prologue,
            emit_module_imports(reduction_callback),
            emit_call(reduction_callback, program, code_filename),
        )
    )


class Reducer:
    """
    Reducer is a wrapper around CReduce.

    It operates on `SourceProgram`s. The reduciont interestingness
    test is implemented as a subclass of `ReductionCallback`.

    Attributes:
        creduce (str): path to the creduce binary
    """

    def __init__(self, creduce: str | None = None):
        """
        Args:
            creduce (str | None):
            path to the creduce binary, if empty "creduce" will be used
        """
        self.creduce = creduce if creduce else "creduce"
        assert which(self.creduce), f"{self.creduce} is not executable"

    def reduce(
        self,
        program: SourceProgram,
        interestingness_test: ReductionCallback,
        jobs: int | None = None,
        log_file: TextIO | None = None,
        debug: bool = False,
    ) -> SourceProgram | None:
        """
        Reduce `program` according to the `interestingness_test`

        Args:
            program (SourceProgram):
                the program to reduce
            interestingness_test (ReductionCallback):
                a concrete ReductionCallback that implementes the interestingness
            jobs (int|None):
                The number of Creduce jobs, if empty cpu_count() will be
            log_file (TextIO | None):
                Where to log Creduce's output, if empty stderr will be used
            debug (bool):
                Whether to pass the debug flag to creduce

        Returns:
            (SourceProgram |None):
                Reduced program, if successful.
        """
        creduce_jobs = jobs if jobs else cpu_count()

        code_filename = "code" + program.language.to_suffix()

        interestingness_script = make_interestingness_script(
            interestingness_test, program, code_filename
        )

        # creduce likes to kill unfinished processes with SIGKILL
        # so they can't clean up after themselves.
        # Setting a temporary temporary directory for creduce to be able to clean
        # up everything
        with TempDirEnv() as tmpdir:
            code_file = tmpdir / code_filename
            with open(code_file, "w") as f:
                f.write(program.code)

            script_path = tmpdir / "check.py"
            with open(script_path, "w") as f:
                print(interestingness_script, file=f)
            os.chmod(script_path, 0o770)
            # run creduce
            creduce_cmd = [
                self.creduce,
                "--n",
                f"{creduce_jobs}",
                str(script_path.name),
                str(code_file.name),
            ]
            if debug:
                creduce_cmd.append("--debug")

            try:
                run_cmd_to_logfile(
                    creduce_cmd,
                    log_file=log_file if log_file else stderr,
                    working_dir=Path(tmpdir),
                    additional_env={"TMPDIR": str(tmpdir.absolute())},
                )
            except subprocess.CalledProcessError as e:
                logging.info(f"Failed to reduce code. Exception: {e}")
                return None

            with open(code_file, "r") as f:
                reduced_code = f.read()

            return replace(program, code=reduced_code)
